#Copyright 2012 Do@. All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification, are
#permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY Do@ ``AS IS'' AND ANY EXPRESS OR IMPLIED
#WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
#CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those of the
#authors and should not be interpreted as representing official policies, either expressed
#or implied, of Do@.

__author__ = 'dvirsky'

import logging
from util import  InstanceCache, Rediston
from patterns.idgenerator import IncrementalIdGenerator


class KeySpec(object):
    """
    Object key descriptor
    """

    def __init__(self, *keys):
        """
        @param keys all the keys to be included in this spec
        """
        self._keys = list(keys)


    def getKey(self, condition):
        """
        Get the proper key for a query
        TODO: Support multiple keys for a single query to be crossed
        @param condition a condition object
        """
        queryKeys = set(condition.fieldsAndValues.iterkeys())
        logging.info("Query keys: %s", queryKeys)

        for key in self._keys:

            if queryKeys == key.fieldSet:
                logging.info("Found key for condition: %s", key)
                return key

        raise ValueError("Could not find key for condition %s", condition)


    def keys(self):
        return self._keys

    @InstanceCache
    def findKeysForUpdate(self, fields):

        fields = set(fields)
        return [key for key  in self._keys if key.fieldSet.intersection(fields)]




class IndexedObject(Rediston):


    #This is the specification of which fields you want to index. override in child classes
    _keySpec = KeySpec({})

    #this is the specification of which fields should be saved to redis
    _spec = ('id',)

    #the id generator for the class. by default it is initialized to an incremental id generator
    #you can replace it with another id generator if you want
    _idGenerator = None


    def __init__(self, **kwargs):
        """
        default constructor, override in subclasses to force strict field typing
        """
        self.id = kwargs.get('id', None)
        self.__dict__.update(kwargs)

        #create id generator the first time needed. this can be overriden in child classes
        if not self.__class__._idGenerator:
            self.__class__._idGenerator = IncrementalIdGenerator(self.__name())


    @classmethod
    def __name(cls):
        """
        the object's name for saving
        """
        if not hasattr(cls, '_oredis_name'):
            setattr(cls, '_oredis_name', cls.__name__.lower())

        return cls._oredis_name

    @classmethod
    def __createId(cls):
        """
       create a new id for an object
        """

        return cls._idGenerator.getId()


    @classmethod
    def __key(cls, id):
        """
        Get the actual id of an object based on a given id
        """
        return '%s:%s' % (cls.__name(), id)

    @classmethod
    def config(cls, host, port, db, timeout = None):

        for k in cls._keySpec.keys():
            k.__class__.config(host, port, db, timeout)

        cls._host = host
        cls._port = port
        cls._db = db
        cls._timeout = timeout


        cls.__connPool = None

    @classmethod
    def loadObjects(cls, ids, *fields):
        """
        Load a list of objects by ids
        @param ids a list of object ids (not keys)
        @param fields optional list of fields to pass if you do not want ALL the object
        """

        p = cls._getPipeline()

        if not fields:

            [p.hgetall(cls.__key(id)) for id in ids]
        else:
            #we get the id anyway,no point in getting it from redis
            if 'id' in fields:
                fields.remove('id')
            [p.hmget(cls.__key(id), fields) for id in ids]

        ret = p.execute()

        objs = []
        for idx,r in enumerate(ret):
            if r:
                r['id'] = ids[idx]
                obj = cls( **r)
                objs.append(obj)
        return objs


    def __repr__(self):

        return '%s(%s)' % (self.__class__.__name__, self.__dict__)

    @classmethod
    def getAll(cls, first = 0, num = -1, *fields):
        """
        Get all the objects of a given type, with optional paging
        """
        redisConn = self._getConnection()
        ids = redisConn.zrange(cls.__classKey(), first, num)
        return cls.loadObjects(ids, redisConn, *fields)




    def __index(self, pipeline = None):
        """
        update the object's indexes
        """

        if not self._keySpec:
            return

        p = pipeline or self._getPipeline('master', transaction=False)
        for k in self._keySpec.keys():
            k.update(self, p)
        if not pipeline:
            p.execute()

    @classmethod
    def __classKey(cls):
        """
        Get the master key of all the available ids in the class
        """
        return 'ids:%s' % cls.__name()

    def __getId(self):
        """
        Get the object's id or create a new one if needed
        """


        if self.id is None:
            self.id = self.__createId()
        return self.id

    def save(self):

        #redisConn = self._getConnection('master')
        _id = self.__getId()

        saveDict = {k: getattr(self, k, None) for k in self._spec}

        pipe =self._getPipeline('master', True)
        #save all properties
        pipe.hmset(self.__key(_id), saveDict )
        #add the id to the master object list
        pipe.zadd(self.__classKey(), **{str(_id): float(_id)})

        #index all the relevant keys
        self.__index(pipe)
        pipe.execute()




    def update(self, **keyValues):
        """
        Set a field(s) in the object and save it to the database
        @param keyValues free form x=y kwargs
        """
        if not self.id:
            raise ValueError("Cannot update a value for an unsaved object")

        #set the data in redis
        ret = self._getConnection('master').hmset(self.__key(self.id), keyValues )

        #update the keys
        updateAbleKeys = self._keySpec.findKeysForUpdate(keyValues.iterkeys())
        for k in updateAbleKeys:
            k.update(self, keyValues)

        #set the data in the object (after successful redis update, to avoid invalid objects)
        for k, v in keyValues:
            setattr(self, k, v)

        return ret


    @classmethod
    def incrementWhere(cls, condition, fieldName, amount):
        """
        Increment a field by an amount, updating the database
        @return a list of updated ids and the new value afterupdate of the field
        """

        ids = cls.find(condition)
        pipe = self._getPipeline('master')
        for id in ids:
            pipe.hincrby(cls.__key(id), fieldName, amount)

        #execute the pipe and get the new values
        newVals = pipe.execute()

        #udpate the keys
        updateAbleKeys = self.__keySpec.findKeysForUpdate((fieldName,))
        for key in updateAbleKeys:
            key.updateMany(((id, newVals[idx]) for idx, id in enumerate(ids)))


        return [(id, newVals[idx]) for idx, id in enumerate(ids)]

    @classmethod
    def updateWhere(condition, **keyValues):
        """
        Update fields with new values for objects matching a condition
        """
        ids = cls.find(condition)
        pipe = self._getPipeline('master')
        #update the database
        for id in ids:
            pipe.hmset(cls.__key(id), keyValues)

        #execute the pipe and get the new values
        res = pipe.execute()

        #udpate the keys
        updateAbleKeys = self.__keySpec.findKeysForUpdate(keyValues.iterkeys())
        for key in updateAbleKeys:
            key.updateMany(((id, keyValues) for ids in ids))

        return ids

    @classmethod
    def get(cls, condition, *fields):
        """
        Load a class by a named key indexing some if its fields
        value can by a multiple token string
        """

        ids = cls.find(condition)

        return cls.loadObjects(ids, *fields )

    @classmethod
    def find(cls, condition):
        """
        Find object ids for a given condition
        """

        key = cls._keySpec.getKey(condition)

        ids = key.find(condition)

        logging.debug("Ids for %s: %s", condition, ids)
        return ids


#Copyright 2012 Do@. All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification, are
#permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY Do@ ``AS IS'' AND ANY EXPRESS OR IMPLIED
#WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
#CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those of the
#authors and should not be interpreted as representing official policies, either expressed
#or implied, of Do@.

__author__ = 'dvirsky'

from util import Rediston, InstanceCache
from threading import RLock
from collections import deque
import logging


class IncrementalIdGenerator(Rediston):
    """
    This class generates guaranteed unique incremental object ids, using redis, that can be pulled from many machines.
    To optimize performance, it doesn't need to request a new id from redis each time.
    Instead it reserves N ids in the clients, and hands them out until they run out.
    This is thread safe and can be used from multiple processes or machines
    """
    def __init__(self, namespace, maxReserveBuffer = 100):
        """
        @param namespace this is the namespace of the ids to be generated. each namespace is a single id generator
        @param maxReserveBuffer how many ids we want to reserve from redis. If you can't accept any "holes" in object
        ids in case of a process crash, set this to 1
        """
        self.namespace = namespace
        self.maxReserveBuffer = maxReserveBuffer
        self.reservedIdsCache = deque()
        self.__lock = RLock()

    @InstanceCache
    def __redisKey(self):
        """
        The redis key for the generator
        """
        return ':%s:idgen' % self.namespace

    def __reserveIds(self):
        """
        Call redis and reserve more ids
        """
        with self.__lock:
            conn = self._getConnection('master')
            res = conn.incr(self.__redisKey(), self.maxReserveBuffer)
            self.reservedIdsCache = deque((res - (self.maxReserveBuffer - i - 1) for i in xrange(self.maxReserveBuffer)))

            #logging.info("Reserved new ids: %s", list(self.reservedIdsCache))

    def getId(self):
        """
        Pop one id, request more from redis if there aren't any left in the cache
        """
        with self.__lock:
            try:
                return self.reservedIdsCache.popleft()
            except IndexError: #queue is empty
                self.__reserveIds()
                return self.reservedIdsCache.popleft()









#Copyright 2012 Do@. All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification, are
#permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY Do@ ``AS IS'' AND ANY EXPRESS OR IMPLIED
#WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
#CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those of the
#authors and should not be interpreted as representing official policies, either expressed
#or implied, of Do@.

import string

import codecs
import re
from util import Rediston, InstanceCache
from patterns.object_store.condition import Condition
import pyhash
import logging


class AbstractKey(object):

    def __init__(self, prefix,fields):

        self.prefix = prefix
        self.fields = fields
        self.fieldSet = set(fields)


    def update(self, obj, pipeline = None):
        pass

    def updateMany(self, idsAndKeyValues = ()):
        pass

    def __repr__(self):

        return '%s(%s:%s)' % (self.__class__.__name__, self.prefix, ','.join(self.fields))


class FullTextKey(AbstractKey, Rediston):
    '''
    classdocs
    '''
    
    trantab = string.maketrans("-_'", "   ")
    stopchars = "\"'\\`'[]{}(),./?:)(*&^%$#@!="

    def __init__(self, prefix, alias, fields, objectScoringCallback = None, delimiter = ' '):
        '''
        Constructor
        '''

        AbstractKey.__init__(self, prefix, fields = [alias,])
        Rediston.__init__(self)
        self.fieldSpec = fields #we don't use the key's "fields" to be able to query multiple fields at once
        self.delimiter = delimiter
        self.scoringCallback = objectScoringCallback

    def getKey(self, word):
        
        return 'ft:%s:%s' % (self.prefix, word)
    
        
    def normalizeString(self, str_):
        
        str_ = codecs.encode(str_, 'utf-8')
        
        return str_.translate(self.trantab, self.stopchars)
        
        
    def update(self, obj, pipeline = None):

        score = 1.0

        #if the object suppports scoring, call the callback now
        if self.scoringCallback:
            score = self.scoringCallback(obj)

        pipe = pipeline or self._getPipeline(transaction=False)
        indexKeys = {}
        #split the words
        for field, factor in self.fieldSpec.iteritems():

            for token in re.split(self.delimiter, getattr(obj, field, '')):
                
                t = self.normalizeString(token.lower().strip())
                
                if t:
                    indexKeys[t]= indexKeys.get(t, 0) + float(factor)*score

        for x in indexKeys:

            pipe.zadd(self.getKey(x), obj.id, indexKeys[x])

        if not pipeline:
            pipe.execute()
            
        
    def find(self, condition):


        string = condition.getValuesFor(self.fields[0])[0]

        tokens = filter(None, (self.normalizeString(value.lower().strip()) for value in re.split(self.delimiter, string)))
        
        
        if not tokens:
            return []
        keys = [self.getKey(t) for t in tokens]
        
        destKey = ('tk:%s' % '|'.join(tokens))
        pipe = self._getPipeline(transaction=False)
        pipe.zinterstore(destKey, keys, 'SUM')
        pipe.zrevrange(destKey, 0, -1, False)
        rx = pipe.execute()
        return rx[1]




class UnorderedKey(AbstractKey, Rediston):
    """
    A simple catch all key, non unique, ideal for short texts (emails, etc). case sensitive.
    it uses hashing of the value as a score in a sorted set
    """
    def __init__(self, prefix, fields):
        '''
        Constructor
        '''
        AbstractKey.__init__(self, prefix, fields)
        Rediston.__init__(self)
        self.hasher = pyhash.fnv1a_64()


    def getValue(self, _dict):


        vals = '::'.join(('%s' % _dict[f] for f in self.fields))

        #make a hash val that is 52 bits and can fit as a sorted set score
        hashval = self.hasher(vals) & 0b11111111111111111111111111111111111111111111111111111

        logging.info("Vals for key %s: %s. hashval: %s", self, vals, hashval)

        return hashval

    @InstanceCache
    def redisKey(self):

        return 'k:%s:%s' % (self.prefix, ','.join(self.fields))

    def update(self, obj, pipeline = None):
        """
        Update the key with the value of this object
        """

        hashval = self.getValue(obj.__dict__)
        conn = pipeline or self._getConnection('master')
        conn.zadd(self.redisKey(), **{str(obj.id): hashval})



    def updateMany(self, ids, cls):

        #first, get the values for these ids
        objs = cls.loadObjects(ids, *self.fields)

        #build a dictionary of all the values to be updated
        updateDict = {}
        for obj in objs:
            if obj:
                updateDict[obj.id] = self.getValue(obj.__dict__)

        conn = self._getConnection('master')
        conn.zadd(self.redisKey(), **updateDict)


    def find(self, condition):
        """
        find objects matching  a certian condition
        currently the condition has to be exactly field=value. no multiple options and no ranges allowed
        """
        hashval = self.getValue(condition.fieldsAndValues)
        conn = self._getConnection()
        return conn.zrangebyscore(self.redisKey(), min=hashval,max=hashval,
                                    start = 0 if not condition.paging else condition.paging[0],
                                    num = -1 if not condition.paging else condition.paging[0] + condition.paging[1] )



class OrderedNumericalKey(AbstractKey, Rediston):
    """
    A key for numerical fields (ints or floats) that can sort and page results
    """
    def __init__(self, prefix, field):
        '''
        Constructor
        '''
        AbstractKey.__init__(self, prefix, (field,))
        Rediston.__init__(self)
        self.field = field


    def getValue(self, _dict):

        return float(_dict[self.field])

    @InstanceCache
    def redisKey(self):

        return 'ok:%s:%s' % (self.prefix, ','.join(self.fields))

    def update(self, obj, pipeline = None):
        """
        Update the key with the value of this object
        """
        val = self.getValue(obj.__dict__)
        conn = pipeline or self._getConnection('master')
        conn.zadd(self.redisKey(), **{str(obj.id): val})



    def updateMany(self, ids, cls):

        #first, get the values for these ids
        objs = cls.loadObjects(ids, self.field)

        #build a dictionary of all the values to be updated
        updateDict = {}
        for obj in objs:
            if obj:
                updateDict[obj.id] = self.getValue(obj.__dict__)

        conn = self._getConnection('master')
        conn.zadd(self.redisKey(), **updateDict)


    def find(self, condition):
        """
        find objects matching  a certian condition
        currently the condition has to be exactly field=value. no multiple options and no ranges allowed
        """

        _min = None
        _max = None
        conditionValue = condition.getValuesFor(self.field)[0]
        #if this is a range query, get the min/max of the query
        if isinstance(conditionValue, Condition.Between):
            _min = float(conditionValue.min)
            _max = float(conditionValue.max)
        elif isinstance(conditionValue, Condition.ConditionType):
            _min = _max = float(conditionValue.value)
        else:
            _min = _max = float(conditionValue)

        conn = self._getConnection()
        return conn.zrangebyscore(self.redisKey(), min=_min,max=_max,
            start = 0 if not condition.paging else condition.paging[0],
            num = -1 if not condition.paging else condition.paging[0] + condition.paging[1] )



class UniqueKeyDuplicateError(Exception):

    pass



class UniqueKey(AbstractKey, Rediston):
    """
    Unique value key, for strings or numbers
    It uses a HASH of value=>objectId to make sure no two objects have the same value
    it raises UniqueKeyDuplicateError if you hit a duplicate value
    """
    def __init__(self, prefix, fields):
        '''
        Constructor
        '''
        AbstractKey.__init__(self, prefix, fields)
        Rediston.__init__(self)


    def getValue(self, _dict):


        return '::'.join(('%s' % _dict[f] for f in self.fields))

    @InstanceCache
    def redisKey(self):

        return 'uk:%s:%s' % (self.prefix, ','.join(self.fields))


    def update(self, obj, pipeline = None):
        """
        Update the key with the value of this object
        """

        val = self.getValue(obj.__dict__)
        pipe = self._getPipeline('master')

        #we set and check the value at once, in case the key is already taken we need to check the current object...
        pipe.hsetnx(self.redisKey(), val, obj.id)
        pipe.hget(self.redisKey(), val)

        rx = pipe.execute()
        #this means
        if rx[0] == 0:
            currentObjId= rx[1]
            logging.info("Possible unique key collision, checking... object: %s, current in key: %s", obj.id, currentObjId)

            if currentObjId != '%s' % obj.id:
                logging.warn("Unique Key collision. wanted to set %s but key already has %s", obj, currentObjId)
                raise UniqueKeyDuplicateError("Duplicate error for key %s" % self)
            else:
                logging.info("Unique key set to the same object")
        else:
            logging.info("Unique key %s set new value for %s:%s", self, obj.id, val)

    def updateMany(self, ids, cls):

        #first, get the values for these ids
        objs = cls.loadObjects(ids, *self.fields)

        #build a dictionary of all the values to be updated
        updateDict = {}
        for obj in objs:
            if obj:
                self.update(obj)




    def find(self, condition):
        """
        find objects matching  a certian condition
        currently the condition has to be exactly field=value. no multiple options and no ranges allowed
        """
        val = self.getValue(condition.fieldsAndValues)
        conn = self._getConnection()
        id = conn.hget(self.redisKey(), val)
        return [id] if id is not None else []




#Copyright 2012 Do@. All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification, are
#permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY Do@ ``AS IS'' AND ANY EXPRESS OR IMPLIED
#WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
#CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those of the
#authors and should not be interpreted as representing official policies, either expressed
#or implied, of Do@.

"""
Various Util functinos
"""
__author__ = 'dvirsky'



'''
Generic singleton wrapper
Created on Sep 21, 2011

@author: dvirsky
'''

import redis


class Rediston(object):
    """
    This is a base class that holds a connection pool and redis connection instances
    """

    __connPool = None

    _host = 'localhost'
    _port = 6379
    _db = 0
    _timeout = None

    @classmethod
    def _getConnection(cls, mode = 'master'):

        if not hasattr(cls, 'redis'):
            cls.redis = None

        if not cls.__connPool:
            cls.__connPool = redis.ConnectionPool(host = cls._host,
                port = cls._port,
                db = cls._db,
                socket_timeout = cls._timeout,
            )

        if not cls.redis:
            cls.redis = redis.Redis(connection_pool = cls.__connPool)

        return cls.redis

    @classmethod
    def _getPipeline(cls,  mode = 'master', transaction = False):
        """
        Create a pipeline object
        @TODO: support master/slave
        """
        conn = cls._getConnection(mode=mode)
        return conn.pipeline(transaction=transaction)



    @classmethod
    def config(cls, host, port, db, timeout = None):

        cls._host = host
        cls._port = port
        cls._db = db
        cls._timeout = timeout


        cls.__connPool = None


    def resetPool(self):
        """
        hard reconnect to avoid forked sockets etc
        """
        self.__class__.__connPool = None
        self.redis = None
        self.__connect()


    def flush(self):

        if getattr(self, 'pipeline', None):
            self.pipeline.execute()



import uuid, base64, struct
def generateRandomId():
    """
    Returns a short, random (see UUID4 RFC), unique id
    """

    i = uuid.uuid4().int
    return base64.b64encode(struct.pack('q', ((i >> 64) ^ i) & 0x7fffffffffffffff) , '-_').strip('=')


def InstanceCache(method):
    """
    Use this to wrap instance methods that return a value that doesn't change after the first calculation
    """

    def wrapped(*args, **kwargs):

        _hash = hash((args[1:], tuple(sorted(kwargs.items()))))
        _self = args[0]
        k = '__MethodCache__%s_%x' % (method.__name__, _hash)

        if hasattr(_self, k):
            return getattr(_self, k)

        ret = method(*args, **kwargs)
        setattr(_self, k, ret)
        return ret

    return wrapped

from contextlib import contextmanager
import time
import sys

@contextmanager
def TimeSampler(actionDescription, callback = None, minTimeFilterMS = 0):

    st = time.time()
    yield
    et = time.time()
    duration =  1000*(et - st)
    if duration < minTimeFilterMS:
        return

    msg = 'Action %s took %.03fms' % (actionDescription, duration)

    if callback:
        callback(msg)
    else:
        sys.stderr.write(msg)



#Copyright 2012 Do@. All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification, are
#permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY Do@ ``AS IS'' AND ANY EXPRESS OR IMPLIED
#WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
#CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those of the
#authors and should not be interpreted as representing official policies, either expressed
#or implied, of Do@.

__author__ = 'dvirsky'

class Condition(object):

    """
    This class represents a condition to fetch objects
    conditions are constructed with a dictionary of field values or ranges, and paging offsets
    """
    class ConditionType(object):
        def __repr__(self):

            return '%s (%s)' % (self.__class__.__name__, self.__dict__)
    """Condition types"""
    class Is(ConditionType):
        """
        Just like ==
        """
        def __init__(self, value):
            self.value = value

    class In(ConditionType):
        """
        Valus is in a multiple option
        """
        def __init__(self, *values):
            raise NotImplementedError("Not implemented yet in any key")
            self.values = values

    class Between(ConditionType):
        """
        Range condition
        """
        def __init__(self, min, max):
            self.min = min
            self.max = max

    def __init__(self, fieldsAndValues, paging = None):
        """
        @param fieldsAndValues a dictionary of fields and their requested values
        @param paging a tuple of (offset, num) to get
        """

        self.fieldsAndValues = fieldsAndValues
        self.paging = paging

    def getValuesFor(self, *fields):

        return [self.fieldsAndValues[f] for f in fields]

    def __repr__(self):

        return 'Condition(%s, paging: %s)' % (self.fieldsAndValues, self.paging)


from distutils.core import setup
setup(name='kickass_redis',
    version='0.1.1',
    package_dir={'kickass_redis': 'src'},
    packages=['kickass_redis', 'kickass_redis.patterns', 'kickass_redis.patterns.object_store'],
)



#Copyright 2012 Do@. All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification, are
#permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY Do@ ``AS IS'' AND ANY EXPRESS OR IMPLIED
#WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
#CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those of the
#authors and should not be interpreted as representing official policies, either expressed
#or implied, of Do@.

__author__ = 'dvirsky'

from util import Rediston, TimeSampler, InstanceCache
import logging
import time

from idgenerator import  IncrementalIdGenerator

class BitmapCounter(Rediston):
    """
    This class wraps a unique id counter, mainly to be used as a user counter
    It can be set to sample specific time resolutions, the default is daily
    It can aggregate several time resolutions into one count, to be cached for memory optimization
    (TBD: the calss should be able to do this)
    """

    RES_DAY = 86400
    RES_HOUR = 3600
    RES_MINUTE = 60

    OP_TOTAL = 'TOTAL'
    OP_AVG = 'AVG'
    OP_INTERESECT = 'INTERSECT'



    def __init__(self, metricName, timeResolutions = (86400,), idMapper = None):
        """
        Constructor
        @param metricName the name of the metric we're sampling, to be used as the redis key
        @param timeResolutions a tuple of resolutions to be sampled, in seconds
        @param idMapper optional IdMapper object that can convert non sequential ids to sequential ones
        NOTE: there will be an extra counter key in redis for each resolution, so lots of resolutions can cause huge RAM overhead
        """
        self.metric = metricName
        self.timeResolutions = timeResolutions
        self.idMapper = idMapper

    def getKey(self, timestamp, resolution = None):
        """
        Get the redis key for this object, for internal use
        """

        resolution = resolution or self.timeResolutions[0]
        return 'uc:%s:%s:%s' % (self.metric, resolution, int(timestamp - (timestamp % resolution)))


    def add(self, objectId, timestamp = None, sequentialIdMappingPrefix = None):
        """
        Add one sample.
        TODO: enable multiple samples in one pipeline
        @param objectId an integer of the object's id. NOTE: do not use this on huge numbers
        @param timestamp the event time, defaults to now
        """

        #map to a sequential id if needed
        if self.idMapper:
            objectId = self.idMapper.getSequentialId(objectId)

        timestamp = timestamp or time.time()

        #get the keys to sample to
        keys = [self.getKey(timestamp, res) for res in self.timeResolutions]
        pipe = self._getPipeline()
        #set the bits
        [pipe.setbit(key, int(objectId), 1) for key in keys]
        pipe.execute()

    def isSet(self, objectId, timestamp, timeResolution = None):
        """
        Tell us whether a specific objectId is set in the counter for a specific resolution
        @param objectId the object to test
        @param timestamp the time to test
        @param timeResolution the time slot to test, defaults to the first resolution given to the counter
        """
        timeResolution = timeResolution or self.timeResolutions[0]
        key = self.getKey(timestamp, timeResolution)
        return self._getConnection().getbit(key, objectId)


    def getCount(self, timestamps, timeResolution = None):
        """
        Count the cardinality of time slots
        @param timestamps a list of timestamps to test
        @param timeResolution the time slot to aggregate, defaults to the first resolution given to the counter
        @return a list of [(timestamp, count), ...]
        """
        timeResolution = timeResolution or self.timeResolutions[0]
        pipe = self._getPipeline()

        [pipe.execute_command('BITCOUNT', self.getKey(timestamp, timeResolution)) for timestamp in timestamps]
        return zip(timestamps, pipe.execute())



    def aggregateCounts(self,  timestamps, op = OP_TOTAL,timeResolution = None, expire = True ):
        """
        Aggregate a few time slots, either summing the unique total, average or memebers in all slots
        @param timestamps a list of timestamps to test
        @param op should be one of SUM, AVG, INTERSECT
        @param timeResolution the time slot to aggregate, defaults to the first resolution given to the counter
        @param expire if set to true, we expire the result
        """
        timeResolution = timeResolution or self.timeResolutions[0]

        if op not in  (BitmapCounter.OP_INTERESECT, BitmapCounter.OP_TOTAL, BitmapCounter.OP_AVG):
            raise ValueError("Invalid aggregation op %s" % op)

        if op == BitmapCounter.OP_INTERESECT:
            bitop = 'AND'
        else:
            bitop = 'OR'

        dest = 'aggregate:%s:%s' % (self.metric,hash(timestamps))
        pipe = self._getPipeline()
        pipe.execute_command('BITOP',bitop, dest, *(self.getKey(timestamp, timeResolution) for timestamp in timestamps))
        pipe.execute_command('BITCOUNT', dest)
        if expire:
            pipe.expire(dest, 60)
        rx = pipe.execute()
        ret = rx[1]
        if op ==  BitmapCounter.OP_AVG:
            return float(ret)/len(timestamps)
        else:
            return ret

    def cohortAnalysis(self, timestamps, timeResolution, filterBitmapKey = None):
        """
        Given a list of timestamps, generates a list of retention measures of the first timestamp, for each later timestamp
        @param timestamps a tuple of timestamps to sample
        @param timeResolution time resolution to sample
        @param filterBitmapKey if set, we intersect it with each sample, to enable selective cohort
        @return a list of tuples [(timestamp,num),...]
        """

        #put the first timestamp as the first record in what we return
        ret = []
        conn = self._getConnection()
        #get the count for each timestamp
        for idx, ts in enumerate(timestamps):
            dest = 'cohort:%s:%s:%s' % (self.metric,ts,idx)
            bitmaps = [self.getKey(timestamps[0], timeResolution), self.getKey(ts, timeResolution)]

            #add filtering bitmap if needed
            if filterBitmapKey:
                bitmaps.append(filterBitmapKey)

            conn.execute_command('BITOP', 'AND', dest, *bitmaps)
            count = conn.execute_command('BITCOUNT', dest)

            ret.append((ts,count))
            conn.expire(dest, 60)


        return ret


    def funnelAnalysis(self, timestamps, timeResolution, filterBitmapKey = None):
        """
        Given a list of timestamps, return a funnel analysis - i.e. for each timestamp, an interesection of it and all the previous points
        @param timestamps a tuple of timestamps to sample
        @param timeResolution time resolution to sample
        @param filterBitmapKey if set, we intersect it with the first sample, to enable selective funnel
        @return a list of tuples [(timestamp,num),...]
        """
        conn = self._getConnection()

        prev = None
        ret = []
        #get the count for each timestamp
        for i in xrange(len(timestamps)):
            dest = 'funnel:%s:%s:%s' % (self.metric,timestamps[i],i)
            conn.execute_command('BITOP', 'AND', dest, prev or filterBitmapKey or self.getKey(timestamps[0], timeResolution),
                                 self.getKey(timestamps[i], timeResolution))

            count = conn.execute_command('BITCOUNT', dest)
            prev = dest
            logging.info("Funnel for timestamp %s: %s", timestamps[i], count)
            ret.append((timestamps[i],count))
            conn.expire(dest, 60)


        return ret



class IdMapper(Rediston):

    """
    This class creates a compact mapping between a non sequential object id to a sequential id
    Create a mapper an pass it to the bitmap counter if you want to convert big or textual ids to sequentials
    """


    def __init__(self, prefix):

        self.prefix = prefix
        self.idgen = IncrementalIdGenerator(namespace=self._redisKey())

    def _redisKey(self):

        return 'idmap:%s' % self.prefix

    def getSequentialId(self, objectId):
        """
        Convert a non sequential id to sequential id, by either creating a new mapping or retrieving an old one
        """

        conn = self._getConnection()
        rc = conn.hget(self._redisKey(), objectId)
        if rc:
            logging.info("Found id mapping for %s:%s: %s", self.prefix, objectId, rc)
            return rc
        else:
            id = self.idgen.getId()
            rc = conn.hsetnx(self._redisKey(), objectId, id)
            if rc: #the write was successful
                logging.info("Created new sequential id for %s:%s: %s", self.prefix, objectId, rc)
                return id
            else: #possible race condition
                logging.info("Got new sequential id for %s:%s: %s", self.prefix, objectId, rc)
                return conn.hget(self._redisKey(), objectId)







